/*

Wiihttpd -- an HTTP server for the Wii

Modified to Wiihttpd by tekencal
Original Source from ftpii v0.0.5

ftpii Source Code Copyright (C) 2008 Joseph Jordan <joe.ftpii@psychlaw.com.au>

This software is provided 'as-is', without any express or implied warranty.
In no event will the authors be held liable for any damages arising from
the use of this software.

Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:

1.The origin of this software must not be misrepresented; you must not
claim that you wrote the original software. If you use this software in a
product, an acknowledgment in the product documentation would be
appreciated but is not required.

2.Altered source versions must be plainly marked as such, and must not be
misrepresented as being the original software.

3.This notice may not be removed or altered from any source distribution.

*/
#include <errno.h>
#include <fat.h>
#include <math.h>
#include <network.h>
#include <ogc/lwp_watchdog.h>
#include <ogcsys.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/dir.h>
#include <unistd.h>
#include <wiiuse/wpad.h>

#include "common.h"

//#include <ogc/ipc.h>
//#include <ogc/system.h>
//#include <gccore.h>
#include <ogc/machine/processor.h>

#define NET_BUFFER_SIZE 1024
#define FREAD_BUFFER_SIZE 1024

const char *CRLF = "\r\n";
const u32 CRLF_LENGTH = 2;

static void *xfb = NULL;
static GXRModeObj *rmode = NULL;

static volatile u8 reset = 0;
static lwp_t reset_thread;

static void reset_called() {
	reset = 1;
}

static void *run_reset_thread(void *arg) {
	while (!reset && !(WPAD_ButtonsHeld(0) & WPAD_BUTTON_A)) {
		sleep(1);
		WPAD_ScanPads();
	}
	printf("\nWiihttpd Webserver shutting down...\n");
	exit(0);
}

u8 initialise_reset_button() {
	s32 result = LWP_CreateThread(&reset_thread, run_reset_thread, NULL, NULL, 0, 80);
	if (result == 0) SYS_SetResetCallback(reset_called);
	return !result;
}

void die(char *msg) {
	perror(msg);
	sleep(5);
	fatUnmount(0);
	//exit(1);
}

void initialise_video() {
    VIDEO_Init();
    rmode = VIDEO_GetPreferredMode(NULL);
    xfb = MEM_K0_TO_K1(SYS_AllocateFramebuffer(rmode));
    console_init(xfb, 20, 20, rmode->fbWidth, rmode->xfbHeight, rmode->fbWidth * VI_DISPLAY_PIX_SZ);
    VIDEO_Configure(rmode);
    VIDEO_SetNextFramebuffer(xfb);
    VIDEO_SetBlack(FALSE);
    VIDEO_Flush();
    VIDEO_WaitVSync(); 
    if (rmode->viTVMode & VI_NON_INTERLACE) VIDEO_WaitVSync();
}

static bool can_open_root_fs() {
    DIR_ITER *root = diropen("/");
    if (root) dirclose(root);
    return (bool)root;
}

void initialise_fat() {
    if (!fatInit(8192, true)) die("Unable to initialise FAT subsystem, exiting");
    // try to open root filesystem - if we don't check here, mkdir crashes later
    if (!can_open_root_fs()) die("Unable to open root filesystem, exiting");
    //if (!fatEnableReadAhead(PI_DEFAULT, 64, 128)) printf("Unable to enable FAT read-ahead caching, speed will suffer...\n");
}

static s32 initialise_network() {
    s32 result;
    while ((result = net_init()) == -EAGAIN);
    return result;
}

static s32 net_ip_top_fd = -1;
static s32 __net_hid = -1;

static inline s32 _net_convert_error(s32 ios_retval)
{
	return ios_retval;
/*	if (ios_retval >= 0) return ios_retval;
	if (ios_retval < -sizeof(_net_error_code_map)
		|| !_net_error_code_map[-ios_retval])
			return NET_UNKNOWN_ERROR_OFFSET + ios_retval;
	return -_net_error_code_map[-ios_retval]; */
}

#define NET_HEAP_SIZE				8192
static char __iptop_fs[] ATTRIBUTE_ALIGN(32) = "/dev/net/ip/top";

//#define DEBUG

#ifdef DEBUG
#define debug_printf printf
#define debug_wait printf
#else
#define debug_printf(...) ;
#define debug_wait(...) ;
#endif

void wait_for_network_initialisation() {
    printf("Waiting for network to initialise...\n");
    if (initialise_network() >= 0) {
        char myIP[16];
        if (if_config(myIP, NULL, NULL, true) < 0) die("Error reading IP address, exiting");
        printf("Network initialised.  Wii IP address: %s\n", myIP);
    } else {
        die("Unable to initialise network, exiting");
    }
    
    // blah
    net_ip_top_fd = _net_convert_error(IOS_Open(__iptop_fs, 0));
    __net_hid = iosCreateHeap(NET_HEAP_SIZE);
    
}

s32 create_server(u16 port) {
    s32 server = net_socket(AF_INET, SOCK_STREAM, IPPROTO_IP);
    if (server < 0) die("Error creating socket, exiting");

    struct sockaddr_in bindAddress;
    memset(&bindAddress, 0, sizeof(bindAddress));
    bindAddress.sin_family = AF_INET;
    bindAddress.sin_port = htons(port);
    bindAddress.sin_addr.s_addr = htonl(INADDR_ANY);

    if (net_bind(server, (struct sockaddr *)&bindAddress, sizeof(bindAddress)) < 0) {
        net_close(server);
        die("Error binding socket");
    }
    if (net_listen(server, 3) < 0) {
        net_close(server);
        die("Error listening on socket");
    }

    return server;
}

s32 accept_peer(s32 server, struct sockaddr_in *addr) {
    s32 peer;
    socklen_t addrlen = sizeof(*addr);
    while ((peer = net_accept(server, (struct sockaddr *)addr, &addrlen)) < 0) {
        //net_close(server);
        die("Error accepting connection");
		
    }
    printf("\nAccepted connection from %s!\n", inet_ntoa(addr->sin_addr));
    return peer;
}

enum {
	IOCTL_SO_ACCEPT	= 1,
	IOCTL_SO_BIND,   
	IOCTL_SO_CLOSE,	
	IOCTL_SO_CONNECT, 
	IOCTL_SO_FCNTL,
	IOCTL_SO_GETPEERNAME, // todo
	IOCTL_SO_GETSOCKNAME, // todo
	IOCTL_SO_GETSOCKOPT,  // todo    8
	IOCTL_SO_SETSOCKOPT,  
	IOCTL_SO_LISTEN,
	IOCTL_SO_POLL,        // todo    b
	IOCTLV_SO_RECVFROM,
	IOCTLV_SO_SENDTO,
	IOCTL_SO_SHUTDOWN,    // todo    e
	IOCTL_SO_SOCKET,
	IOCTL_SO_GETHOSTID,
	IOCTL_SO_GETHOSTBYNAME,
	IOCTL_SO_GETHOSTBYADDR,// todo
	IOCTLV_SO_GETNAMEINFO, // todo   13
	IOCTL_SO_UNK14,        // todo
	IOCTL_SO_INETATON,     // todo
	IOCTL_SO_INETPTON,     // todo
	IOCTL_SO_INETNTOP,     // todo
	IOCTLV_SO_GETADDRINFO, // todo
	IOCTL_SO_SOCKATMARK,   // todo
	IOCTLV_SO_UNK1A,       // todo
	IOCTLV_SO_UNK1B,       // todo
	IOCTLV_SO_GETINTERFACEOPT, // todo
	IOCTLV_SO_SETINTERFACEOPT, // todo
	IOCTL_SO_SETINTERFACE,     // todo
	IOCTL_SO_STARTUP,           // 0x1f
	IOCTL_SO_ICMPSOCKET =	0x30, // todo
	IOCTLV_SO_ICMPPING,         // todo
	IOCTL_SO_ICMPCANCEL,        // todo
	IOCTL_SO_ICMPCLOSE          // todo
};

struct bind_params {
	u32 socket;
	u32 has_name;
	u8 name[28];
};

struct sendto_params {
	u32 socket;
	u32 flags;
	u32 has_destaddr;
	u8 destaddr[28];
}; 

static inline s32 my_net_sendto(s32 s, const void *data, s32 len, u32 flags, struct sockaddr *to, socklen_t tolen)
{
	s32 ret;
    u8 * message_buf = NULL;
	STACK_ALIGN(struct sendto_params,params,1,32);
    
	if (net_ip_top_fd < 0) return -ENXIO;
	if (tolen > 28) return -EOVERFLOW;

	message_buf = iosAlloc(__net_hid, len);
	if (message_buf == NULL) {
		debug_printf("net_send: failed to alloc %d bytes\n", len);
		return IPC_ENOMEM;
	}
        
	debug_printf("net_sendto(%d, %p, %d, %d, %p, %d)\n", s, data, len, flags, to, tolen);
/*
	if (to && to->sa_len != tolen) {
		debug_printf("warning: to->sa_len was %d, setting to %d\n",	to->sa_len, tolen);
		to->sa_len = tolen;
	}
*/	
	memset(params, 0, sizeof(struct sendto_params));
	memcpy(message_buf, data, len);   // ensure message buf is aligned

	params->socket = s;  
	params->flags = flags;
    /*
	if (to) {
		params->has_destaddr = 1;
		memcpy(params->destaddr, to, to->sa_len);		
	} else {
        */
		params->has_destaddr = 0;
	//}

	ret = _net_convert_error(IOS_IoctlvFormat(__net_hid, net_ip_top_fd, IOCTLV_SO_SENDTO, "dd:", message_buf, len, params, sizeof(struct sendto_params)));
	debug_printf("net_send retuned %d\n", ret);

	if(message_buf!=NULL) iosFree(__net_hid,message_buf);
	return ret;
}

static inline s32 my_net_write(s32 s, const void *data, s32 size)
{
    return my_net_sendto(s, data, size, 0, NULL, 0);
}

typedef s32 (*transferrer_type)(s32 s, void *mem, s32 len);
inline static s32 transfer_exact(s32 s, char *buf, s32 length, transferrer_type transferrer) {
    s32 bytes_transferred = 0;
    s32 remaining = length;
    while (remaining) {
        if ((bytes_transferred = transferrer(s, buf, remaining > NET_BUFFER_SIZE ? NET_BUFFER_SIZE : remaining)) > 0) {
            remaining -= bytes_transferred;
            buf += bytes_transferred;
        } else if (bytes_transferred < 0) {
            return bytes_transferred;
        } else {
            return -ENODATA;
        }
    }
    return 0;
}

inline s32 write_exact(s32 s, char *buf, s32 length) {
    return transfer_exact(s, buf, length, (transferrer_type)my_net_write);
}

// Read from a file
s32 write_from_file(s32 s, FILE *f) {
    char buf[FREAD_BUFFER_SIZE];
    s32 bytes_read, bytes_written;
    while (1) {
        bytes_read = fread(buf, 1, FREAD_BUFFER_SIZE, f);
        if (bytes_read > 0) {
            if ((bytes_written = write_exact(s, buf, bytes_read)) < 0) {
                printf("DEBUG: write_from_file() net_write error: [%i] %s\n", -bytes_written, strerror(-bytes_written));
                return bytes_written;
            }
        }
        if (bytes_read < FREAD_BUFFER_SIZE) {
            s32 result = -!feof(f);
            if (result < 0) {
                printf("DEBUG: write_from_file() fread error: [%i] %s\n", ferror(f), strerror(ferror(f)));
            }
            return result;
        }
    }
}


// Result must be able to hold up to maxsplit+1 null-terminated strings of length strlen(s)
// Returns the number of strings stored in the result array (up to maxsplit+1)
u32 split(char *s, char sep, u32 maxsplit, char *result[]) {
    u32 num_results = 0;
    u32 result_pos = 0;
    u32 trim_pos = 0;
    bool in_word = false;
    for (; *s; s++) {
        if (*s == sep) {
            if (num_results <= maxsplit) {
                in_word = false;
                continue;
            } else if (!trim_pos) {
                trim_pos = result_pos;
            }
        } else if (trim_pos) {
            trim_pos = 0;
        }
        if (!in_word) {
            in_word = true;
            if (num_results <= maxsplit) {
                num_results++;
                result_pos = 0;
            }
        }
        result[num_results - 1][result_pos++] = *s;
        result[num_results - 1][result_pos] = '\0';
    }
    if (trim_pos) {
        result[num_results - 1][trim_pos] = '\0';
    }
    u32 i = num_results;
    for (i = num_results; i <= maxsplit; i++) {
        result[i][0] = '\0';
    }
    return num_results;
}
